/*								     HTThread.c
**	MULTIPLE THREAD SOCKET MANAGEMENT
**
*/

/* Implemention dependent include files */
#include "tcp.h"

/* Library include files */
#include "HTUtils.h"
#include "HTAccess.h"
#include "HTError.h"
#include "HTThread.h"					 /* Implemented here */
#include "HTEvent.h"
#include "tk.h"

/* Type definitions and global variables etc. local to this module */

PRIVATE fd_set	HTfd_intr;	        /* All sockets currently interrupted */
PRIVATE fd_set	HTfd_set;	         /* All sockets currently registered */
PRIVATE int 	HTMaxfdpl = 0;			/* Max number of sockets + 1 */

PRIVATE HTList *HTThreads = NULL;      	 /* List of the HTNetInfo structures */

/* ------------------------------------------------------------------------- */

/*								   HTThreadInit
**
**	Initiates the thread socket registers. It is very important that
**	this function is called. It is currently done inside HTAccess in the
**	HTAccessInit function.
*/

PRIVATE void HTThreadProcessTkEvent(ClientData user_request, int mask) 
{
  int status = 0;
  HTEventState state;
  HTRequest *cur_request = (HTRequest *) user_request;
  HTMethod method = cur_request->method;

  HTProtocol *callback = (HTProtocol *)
    HTAnchor_protocol(cur_request->anchor);
  if (method == METHOD_GET || method == METHOD_HEAD) {
    status = (*callback->load)(cur_request);
  } else {
    if (THD_TRACE)
      fprintf(stderr, "HTEventLoop. Method not supported @@@\n");
  }
  if (THD_TRACE)
    fprintf(stderr, "HTEventLoop. Load returns: %d\n", status);
  if (status != HT_WOULD_BLOCK) {
    HTLoadTerminate(cur_request, status);
    if (THD_TRACE)
      fprintf(stderr,"HTEventLoop. Calling Terminate\n");
  }
}

PUBLIC BOOL HTThreadInit NOARGS
{
    static BOOL done=NO;

    if (done) {
	if (THD_TRACE)
	    fprintf(stderr, "ThreadInit.. Already done\n");
	return NO;
    }
    done = YES;

    /* Initialize the other internal bit arrays */
    FD_ZERO(&HTfd_set);
    FD_ZERO(&HTfd_intr);
    return YES;
}


/*							      HTThreadGetFDInfo
**
**	Returns the maximum bit width and the read and write bit array.
*/
PUBLIC int HTThreadGetFDInfo ARGS2(fd_set *, read, fd_set *, write)
{
  return 0;
}


/*								  HTThreadState
**
**	This function registers a socket as waiting for the action given
**	(read or write etc.).
*/
PUBLIC void HTThreadState ARGS2(int, sockfd, HTThreadAction, action)
{
  HTNetInfo *pres;
  int mask;
  if (THD_TRACE) {
    static char *actionmsg[] = {
      "SET WRITE",
      "CLEAR WRITE",
      "SET READ",
      "CLEAR READ",
      "SET INTERRUPT",
      "CLEAR INTERRUPT",
      "CLOSE"
      };
    fprintf(stderr,
	    "Thread...... Register socket number %d for action %s\n",
	    sockfd, *(actionmsg+action));
  }

    /* Find the corresponding HTNetInfo structure */
  while ((pres = (HTNetInfo *) HTList_nextObject(HTThreads)) != NULL)
    if (pres->sockfd == sockfd) 
      break;
  
  if (!pres) 
    return;

  switch (action) {
  case THD_SET_WRITE:
    Tk_CreateFileHandler(sockfd, TK_WRITABLE, 
			 HTThreadProcessTkEvent, (ClientData) pres->request);
    FD_SET(sockfd, &HTfd_set);
    break;

  case THD_CLR_WRITE:
    Tk_DeleteFileHandler(sockfd);
    FD_CLR(sockfd, &HTfd_set);
    break;

  case THD_SET_READ:
    Tk_CreateFileHandler(sockfd, TK_READABLE, 
			 HTThreadProcessTkEvent, (ClientData) pres->request);
    FD_SET(sockfd, &HTfd_set);
    break;

  case THD_CLR_READ:
    Tk_DeleteFileHandler(sockfd);
    FD_CLR(sockfd, &HTfd_set);
    break;

  case THD_CLOSE:
    Tk_DeleteFileHandler(sockfd);
    FD_CLR(sockfd, &HTfd_intr);
    FD_CLR(sockfd, &HTfd_set);
    break;

  case THD_SET_INTR:
    Tk_DeleteFileHandler(sockfd);
    FD_SET(sockfd, &HTfd_intr);
    break;

  case THD_CLR_INTR:
    FD_CLR(sockfd, &HTfd_intr);
    break;

  default:
    if (THD_TRACE)
      fprintf(stderr, "Thread...... Illigal socket action\n");
  }
  if (action == THD_CLOSE) {
    if (sockfd+1 >= HTMaxfdpl) {
      while (HTMaxfdpl > 0 && !FD_ISSET(HTMaxfdpl-1, &HTfd_set))
	HTMaxfdpl--;
    }
  } else {
    if (sockfd+1 > HTMaxfdpl)
      HTMaxfdpl = sockfd+1;
  }
  if (THD_TRACE)
    fprintf(stderr, "Thread...... Max bitwidth is %d\n", HTMaxfdpl);
}

/*								   HTThreadIntr
**
**	This function returns YES or NO to the question
*/
PUBLIC BOOL HTThreadIntr ARGS1(int, sockfd)
{
    return FD_ISSET(sockfd, &HTfd_intr) ? YES : NO;
}


/*							    HTThreadMarkIntrAll
**
**	Marks all Library sockets as interrupted. User sockets can not be
**	interrupted
*/
PUBLIC void HTThreadMarkIntrAll ARGS1(CONST fd_set *,	fd_user)
{
    int cnt;
    if (THD_TRACE)
	fprintf(stderr, "Thread...... Mark ALL Library sockfd INTERRUPTED\n");
    for (cnt=0; cnt<HTMaxfdpl; cnt++) {
	if (FD_ISSET(cnt, &HTfd_set) && !FD_ISSET(cnt, fd_user))
	    FD_SET(cnt, &HTfd_intr);
    }
}


/*								 HTThreadActive
**
**	Returns yes as long as a socket other than stdin is registered in 
**	the total set of sockets.
*/
PUBLIC BOOL HTThreadActive NOARGS
{
    int cnt;
    for (cnt=STDIN_FILENO+1; cnt<HTMaxfdpl; cnt++)
	if (FD_ISSET(cnt, &HTfd_set))
	    return YES;
    return NO;
}


/*							     	   HTThread_new
**
**	Register the HTNetInfo structure in a list so that we can find the 
**	request which corresponds to a socket descriptor
*/
PUBLIC void HTThread_new ARGS1(HTNetInfo *, new_net)
{
    if (!HTThreads)
	HTThreads = HTList_new();
    HTList_addObject(HTThreads, (void *) new_net);
}


/*							         HTThread_clear
**
**	Remove the HTNetInfo from the list of acrive threads.
*/
PUBLIC int HTThread_clear ARGS1(HTNetInfo *, old_net)
{
    if (HTThreads)
	return HTList_removeObject(HTThreads, (void *) old_net);
    return NO;
}


/*							   HTThread_getCallBack
**
**	Finds a socket and makes it the active thread. READ has a higher
**	priority than WRITE as it might influence (stop) the WRITE process.
**	Returns the request structure of the active thread, NULL on error.
**	A interrupted socket has the highest priority
*/
PUBLIC HTRequest *HTThread_getRequest ARGS2(CONST fd_set *,	fd_read,
					    CONST fd_set *, 	fd_write)
{
    return NULL;
}
