# expand-mode
#
# When display area is bigger than request area, what to do?
#   adjust : keep the w:window to be the reqwidth/height
#   expand : expand w:window to fill the empty area
#   adjustexpand : first adjust, then expand
#
# Terminology:
#   "child window"
#       - the window that is being scrolled
#   "display window"
#       - the part of the child window that appears on the screen.
#       - i.e., the internal subwidget "int0"
#

tixWidgetClass tixScrolledWindow {
    -superclass tixPrimitive
    -classname  TixScrolledWindow
    -shortname  tixSWin
    -method {
	setsize
    }
    -flag {
	-command -expandmode -scrollbar -reqwidth -reqheight
    }
    -static {
	-scrollbar
    }
    -configspec {
	{-command command Command {}}
	{-expandmode expandMode ExpandMode adjust}
	{-scrollbar scrollbar Scrollbar both}
	{-reqwidth reqWidth ReqWidth 0}
	{-reqheight reqHeight ReqHeight 0}
    }
}

proc tixSWin::InitWidgetRec {w} {
    upvar #0 $w data

    tixChainMethod $w InitWidgetRec

    set data(width)        0
    set data(height)       0
    set data(xview)        0
    set data(yview)        0
    set data(actualwidth)  0
    set data(actualheight) 0

    set data(type)         nil
    set data(mapped)       0

    set data(xserial)      0
    set data(yserial)      0
}

proc tixSWin::ConstructWidget {w} {
    upvar #0 $w data

    tixChainMethod $w ConstructWidget

    # Top frame
    frame $w.top -border 0

    # Frame inside top frame
    frame $w.int0 -border 2 -relief sunken
    frame $w.int1 -border 0
    pack $w.int1 -in $w.int0 -expand yes -fill both

    # Bottom frame
    frame $w.bot -border 0

    scrollbar $w.hsb \
	-orient horizontal \
	-command "tixSWin::View $w x xview width"

    scrollbar $w.vsb \
	-command "tixSWin::View $w y yview height"

    set scbWidth [lindex [$w.vsb config -width] 4]
    set scbBd    [lindex [$w.vsb config -bd]    4]

    frame $w.xframe -height $scbWidth -relief flat \
	-width [expr "$scbWidth + 2 * $scbBd"]

    set data(w_top)     $w.top
    set data(w_bot)     $w.bot
    set data(w_int0)    $w.int0
    set data(w_int1)    $w.int1
    set data(w:hsb)     $w.hsb
    set data(w:vsb)     $w.vsb
    set data(w:xframe)  $w.xframe
    set data(w:window) [tixCallMethod $w ConstructScrolledWindow $data(w_int1)]

    place $data(w:window) -in $data(w_int1) -x 0 -y 0 \
	-bordermode ignore
    tixSWin::Repack $w
}

proc tixSWin::SetBindings {w} {
    upvar #0 $w data

    tixChainMethod $w SetBindings

    bind $data(w_int1)   <Configure> "tixSWin::MonitorSize $w"
    bind $data(w:window) <Map>       "tixSWin::Map $w"
}

proc tixSWin::ConstructScrolledWindow {w parent} {
    upvar #0 $w data

    set f [frame $parent.window]
    return $f
}

proc tixSWin::Repack {w} {
    upvar #0 $w data

    if {$data(-scrollbar) == "auto"} {
	set type [tixSWin::ScrollBarsNeeded $w]
    } else {
	set type $data(-scrollbar)
    }

    if {$type != $data(type)} {
	pack forget $data(w_top)
	pack forget $data(w_bot)
	pack forget $data(w_int0)
	pack forget $data(w:hsb)
	pack forget $data(w:vsb)
	pack forget $data(w:xframe)

	set data(type) $type
	tixSWin::Pack $w
    }
}

proc tixSWin::Pack {w} {
    upvar #0 $w data

    case $data(type) {
	both {
	   pack $data(w:vsb)  -in $data(w_top) -side right -fill y
	   pack $data(w_int0) -in $data(w_top) -side left  -fill both -expand 1

	   pack $data(w:hsb)    -in $data(w_bot) -side left -fill x -expand 1
	   pack $data(w:xframe) -in $data(w_bot) -side left -fill y

	   pack $data(w_top) -side top -fill both -expand yes
	   pack $data(w_bot) -side bottom -fill x
	}
	x {
	    pack $data(w:hsb) -side bottom -fill x
	    pack $data(w_int0) -side top -fill both -expand yes
	}
	y {
	    pack $data(w:vsb) -side right -fill y
	    pack $data(w_int0) -side left -fill both -expand yes
	}
	none {
	    pack $data(w_int0) -fill both -expand yes
	}
    }
}

proc tixSWin::ScrollBarsNeeded {w} {
    upvar #0 $w data

    set pw [winfo width  $data(w_int1)]
    set ph [winfo height $data(w_int1)]

    if {$data(xview) == 0} {
	set ww $data(-reqwidth)
    } else {
	set ww $data(actualwidth)
    }
    if {$data(yview) == 0} {
	set wh $data(-reqheight)
    } else {
	set wh $data(actualheight)
    }

    if {$ww > $pw} {
	if {$wh > $ph} {
	    return both
	} else {
	    return x
	}
    } elseif {$wh > $ph} {
	return y
    } else {
	return none
    }
}

proc tixSWin::CalculateConstraint {w spec view lspec} {
    upvar #0 $w data

     if {[expr "$data($lspec) + $data($view)"] > $data(actual$lspec)} {
	case $data(-expandmode) {
	    adjust {
		set data($view) [expr "$data(actual$lspec) - $data($lspec)"]
		if {$data($view) < 0} {
		    set data($view) 0
		}
		place $data(w:window) -$spec -$data($view) -bordermode ignore
	    }
	    adjustexpand {
		set data($view) [expr "$data(actual$lspec) - $data($lspec)"]
		if {$data($view) <= 0} {
		    set data($view) 0
		    if {$data($lspec) > $data(actual$lspec)} {
			set data(actual$lspec) $data($lspec)
		    }
		}
		place $data(w:window) -$spec -$data($view) \
		    -bordermode ignore -$lspec $data(actual$lspec)
	    }
	    expand {
		set data(actual$lspec) [expr "$data($lspec) + $data($view)"] 
		place $data(w:window) -$spec -$data($view) \
		    -bordermode ignore -$lspec $data(actual$lspec)
	    }
	}
    } else {
	case $data(-expandmode) {
	    {adjustexpand expand} {
		set data(actual$lspec) [expr "$data($lspec) + $data($view)"]
		if {$data(actual$lspec) < $data(-req$lspec)} {
		    set data(actual$lspec) $data(-req$lspec)
		}
		place $data(w:window) -$spec -$data($view) \
		    -bordermode ignore  -$lspec $data(actual$lspec)
	    }
	}
    }
}

# Whenever the display window or the child window change size, this
# function is called to set the scrollbars properly
#
proc tixSWin::MonitorSize {w} {
    upvar #0 $w data

    tixSWin::Repack $w

    set data(width)        [winfo width  $data(w_int1)]
    set data(height)       [winfo height $data(w_int1)]
    set data(actualwidth)  [winfo reqwidth  $data(w:window)]
    set data(actualheight) [winfo reqheight $data(w:window)]

    if {$data(actualwidth) < $data(-reqwidth)} {
	set data(actualwidth) $data(-reqwidth)
    }
    if {$data(actualheight) < $data(-reqheight)} {
	set data(actualheight) $data(-reqheight)
    }

    set data(xview) [expr "- [winfo x $data(w:window)]"]
    set data(yview) [expr "- [winfo y $data(w:window)]"]

    tixSWin::CalculateConstraint $w x xview width
    tixSWin::CalculateConstraint $w y yview height

    tixSWin::SetScrollBars $w
}

proc tixSWin::SetScrollBars {w} {
    upvar #0 $w data

    set xfirst [expr "$data(xview)"]
    set yfirst [expr "$data(yview)"]
    set xlast  [expr "$xfirst + $data(width)"]
    set ylast  [expr "$yfirst + $data(height)"]

    if {$xlast > $data(actualwidth)} {
	set xlast $data(actualwidth)
    }
    if {$ylast > $data(actualheight)} {
	set ylast $data(actualheight)
    }

    $data(w:hsb) set $data(actualwidth)  $data(width)  $xfirst $xlast
    $data(w:vsb) set $data(actualheight) $data(height) $yfirst $ylast
}

proc tixSWin::View {w spec view lspec val} {
    upvar #0 $w data

    set sflag [format "%sserial" $spec]
    incr data($sflag)
    after 0 tixSWin::ViewCompressed $w $spec $view $lspec $val $data($sflag)
}

proc tixSWin::ViewCompressed {w spec view lspec val serial} {
    upvar #0 $w data

    set sflag [format "%sserial" $spec]

    if {$serial < $data($sflag)} {
	return
    }

    if {$val < 0} {
	set val 0
    } elseif { [expr "$data($lspec) + $val"] > $data(actual$lspec)} {
	set val [expr "$data(actual$lspec) - $data($lspec)"]
    }

    place $data(w:window) -$spec -$val -bordermode ignore \
	-$lspec $data(actual$lspec)

    set data($view) $val

    tixSWin::SetScrollBars $w

    if {$val == 0 && $data(-expandmode) == "expand"} {
	if {$data($lspec) > $data(-req$lspec)} {
	    after 300 tixSWin::Repack $w
	}
    }
}

# setsize
#
# The scrolled area changes its size. Notifies the scrolledwindow, which
# changes the scrollbars automatically
#
proc tixSWin::config-reqwidth {w value} {
    upvar #0 $w data

    if {$value > $data(actualwidth)} {
	set data(actualwidth) $value
    }
    after 0 tixSWin::MonitorSize $w
}

proc tixSWin::config-reqheight {w value} {
    upvar #0 $w data

    if {$value > $data(actualheight)} {
	set data(actualheight) $value
    }
    after 0 tixSWin::MonitorSize $w
}


proc tixSWin::Map {w} {
    upvar #0 $w data

    if {$data(mapped) == 0 && $data(-reqwidth) == 0} {
	update idletasks
	set data(-reqwidth)  [winfo reqwidth  $data(w:window)]
	set data(-reqheight) [winfo reqheight $data(w:window)]
	after 0 tixSWin::MonitorSize $w
	set data(mapped) 1
    }
}
